import sys
import random
import math

import pygame
import const
import util
import data


class GameWindow(object):
    
    def __init__(self):
        pygame.display.set_caption('Moonrays')
        self.screen = pygame.display.set_mode((2*const.WIDTH, 2*const.HEIGHT))
        try:
            pygame.mixer.init()
            pygame.mixer.music.load(data.filepath('Plaint.mp3'))
            pygame.mixer.music.play(-1)
        except:
            print 'Cannot load sound. :('
        self.intro()
        # self.game()

    def intro(self):
        Intro(self, self.screen).loop()
        self.game()

    def game(self):
        Game(self, self.screen).loop()


class Intro(object):

    def __init__(self, window, screen):
        self.window = window
        self.real_screen = screen
        self.screen = pygame.surface.Surface((const.WIDTH, const.HEIGHT))
        pygame.font.init()
        self.font = pygame.font.Font(data.filepath('Qlassik_TB.ttf'), 24)
        self.font_16 = pygame.font.Font(data.filepath('Qlassik_TB.ttf'), 14)
        text = self.font.render('Moonrays', False, (192, 192, 255))
        self.screen.blit(text, (170, 30))
        text = self.font_16.render('A creature is hunting you. Find a way to use moonlight to track it', False, (192, 192, 255))
        self.screen.blit(text, (20, 120))
        text = self.font_16.render('and shoot it before it catches you. You cannot shoot while it sees you.', False, (192, 192, 255))
        self.screen.blit(text, (20, 140))
        text = self.font_16.render('Instructions:', False, (192, 192, 255))
        self.screen.blit(text, (20, 180))
        text = self.font_16.render('X + arrow key to place mirror (only on moonlight or rays)', False, (192, 192, 255))
        self.screen.blit(text, (20, 200))
        text = self.font_16.render('C + arrow key to shoot', False, (192, 192, 255))
        self.screen.blit(text, (20, 220))
        text = self.font_16.render('Good luck!', False, (192, 192, 255))
        self.screen.blit(text, (20, 270))
        pygame.transform.scale(self.screen, (2*const.WIDTH, 2*const.HEIGHT), self.real_screen)
        pygame.display.update()

    def loop(self):
        self.done = False
        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        sys.exit()
                    else:
                        self.done = True


class Game(object):
    
    def __init__(self, window, screen):
        self.window = window
        self.real_screen = screen
        self.screen = pygame.surface.Surface((const.WIDTH, const.HEIGHT))
        
        self.clock = pygame.time.Clock()
        
        pygame.font.init()
        
        self.paused = False
        self.invul_timer = 0
        
        self.bullet_mgr = BulletManager()
        self.player = Player((100,100), self.bullet_mgr)
        self.player_con = pygame.sprite.RenderUpdates(self.player)
        self.tilemap = Tilemap()
        self.tilemap.add_light(TileMoonlight((4, 5)))
        self.tilemap.add_light(TileMoonlight((7, 15)))
        self.tilemap.add_light(TileMoonlight((13, 9)))
        for i in range(const.TILES_Y):
            for j in range(const.TILES_X):
                self.tilemap.add_tile(TileGrass((i, j)))
        
        self.damage_screen_con = pygame.sprite.RenderUpdates()
        self.damage_screen = DamageScreen()
        self.hunted_screen = HuntedScreen()
        self.hunted_screen_con = pygame.sprite.GroupSingle(self.hunted_screen)
        
        self.enemies = pygame.sprite.RenderUpdates()
        # self.enemies.add(Enemy((100,100), self.player))
        self.enemies.add(Enemy((200,200), self.player))
        
        self.rayblockmgr = RayBlockManager(self.player, self.enemies, self.tilemap)

        self.sight = Sight(self.player)
        self.sight_con = pygame.sprite.RenderUpdates(self.sight)

        self.font = pygame.font.Font(data.filepath('Qlassik_TB.ttf'), 24)
        self.font_10 = pygame.font.Font(data.filepath('Qlassik_TB.ttf'), 10)
        
        self.arrows = 5
    
    def loop(self):
        self.bg = pygame.Surface((const.WIDTH, const.HEIGHT))
        self.bg.fill((0,0,0))
        self.screen.blit(self.bg, (0,0))
        
        pygame.transform.scale(self.screen, (2*const.WIDTH, 2*const.HEIGHT), self.real_screen)
        pygame.display.update()
        
        self.done = False
        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    self.on_keydown(event)
                elif event.type == pygame.KEYUP:
                    self.on_keyup(event)
            
            if not self.paused:
                self.tilemap.map_sprites.clear(self.screen, self.bg)
                self.enemies.clear(self.screen, self.bg)
                self.sight_con.clear(self.screen, self.bg)
                self.tilemap.light_sprites.clear(self.screen, self.bg)
                self.tilemap.object_sprites.clear(self.screen, self.bg)
                self.player_con.clear(self.screen, self.bg)
                self.bullet_mgr.clear(self.screen, self.bg)
                for mirror in self.tilemap.mirrors:
                    if mirror.ray is not None: mirror.ray.clear(self.screen, self.bg)
                self.hunted_screen_con.clear(self.screen, self.bg)
                self.damage_screen_con.clear(self.screen, self.bg)
            
            dt = self.clock.tick(const.FPS) / 1000.0
            
            if not self.paused:
                self.update(dt)
                self.rayblockmgr.update(dt)
                self.tilemap.update(dt)
                self.player.update(dt)
                for enemy in self.enemies:
                    enemy.update(dt)
                self.sight_con.update(dt)
                self.bullet_mgr.update(dt)
                for mirror in self.tilemap.mirrors:
                    if mirror.ray is not None: mirror.ray.update(dt)
                self.hunted_screen.update(dt)
                self.damage_screen.update(dt)
                if not self.damage_screen.active:
                    self.damage_screen_con.empty()
            
            if not self.paused:
                self.tilemap.map_sprites.draw(self.screen)
                self.enemies.draw(self.screen)
                self.sight_con.draw(self.screen)
                self.tilemap.light_sprites.draw(self.screen)
                self.tilemap.object_sprites.draw(self.screen)
                self.player_con.draw(self.screen)
                self.bullet_mgr.draw(self.screen)
                for mirror in self.tilemap.mirrors:
                    if mirror.ray is not None: mirror.ray.draw(self.screen)
                self.hunted_screen_con.draw(self.screen)
                self.damage_screen_con.draw(self.screen)

                text = self.font_10.render(str(self.arrows) + ' arrows', False, (192, 192, 255))
                self.screen.blit(text, (10, 10))
            
            pygame.transform.scale(self.screen, (2*const.WIDTH, 2*const.HEIGHT), self.real_screen)
            pygame.display.flip()
    
    def pause(self):
        self.paused = not self.paused
    
    def on_keydown(self, event):
        if event.key == pygame.K_ESCAPE:
            if self.paused:
                sys.exit()
            else:
                self.pause()
        elif event.key == pygame.K_p:
            self.pause()
        elif event.key == pygame.K_LEFT or event.key == pygame.K_a:
            self.player.move_left()
        elif event.key == pygame.K_RIGHT or event.key == pygame.K_d:
            self.player.move_right()
        elif event.key == pygame.K_UP or event.key == pygame.K_w:
            self.player.move_up()
        elif event.key == pygame.K_DOWN or event.key == pygame.K_s:
            self.player.move_down()
        elif event.key == pygame.K_x:
            self.player.start_directed_action(False)
        elif event.key == pygame.K_c:
            if not self.hunted_screen.hunted:
                self.player.start_directed_action(True)

    def on_keyup(self, event):
        if event.key == pygame.K_LEFT or event.key == pygame.K_a:
            self.player.stop_move_left()
        elif event.key == pygame.K_RIGHT or event.key == pygame.K_d:
            self.player.stop_move_right()
        elif event.key == pygame.K_UP or event.key == pygame.K_w:
            self.player.stop_move_up()
        elif event.key == pygame.K_DOWN or event.key == pygame.K_s:
            self.player.stop_move_down()
        elif event.key == pygame.K_x:
            self.player.place_mirror_done(self.tilemap)
        elif event.key == pygame.K_c:
            if self.arrows > 0:
                if self.player.shoot_done(self.hunted_screen.hunted):
                    self.arrows -= 1
    
    def damage(self):
        self.damage_screen_con.add(self.damage_screen)
        self.damage_screen.start(self.die)
    
    def die(self):
        self.paused = True
        text = self.font.render('You are dead!', False, (255, 24, 24))
        self.screen.blit(text, (150, 140))
        text = self.font_10.render('Press ESC to quit the game', False, (255, 255, 255))
        self.screen.blit(text, (10, 280))
    
    def die_arrows(self):
        self.paused = True
        text = self.font.render('You are out of arrows...', False, (255, 24, 24))
        self.screen.blit(text, (70, 140))
        text = self.font_10.render('Press ESC to quit the game', False, (255, 255, 255))
        self.screen.blit(text, (10, 280))
    
    def win(self):
        self.paused = True
        text = self.font.render('You have killed it! Congratulations!', False, (192, 192, 255))
        self.screen.blit(text, (50, 140))
        text = self.font_10.render('Press ESC to quit the game', False, (255, 255, 255))
        self.screen.blit(text, (10, 280))

    def update(self, dt):
        self.invul_timer -= dt
        if self.invul_timer <= 0:
            self.invul_timer = 0
            for enemy in self.enemies:
                if pygame.Rect.colliderect(pygame.Rect(enemy.rect.left + 2, enemy.rect.top + 2, enemy.rect.width - 4, enemy.rect.height - 4), self.player.get_bb()):
                    self.damage()
                    self.invul_timer = 2
        self.hunted_screen.set_active(False)
        for enemy in self.enemies:
            if enemy.state == const.STATE_HUNT:
                self.hunted_screen.set_active(True)
            for bullet in self.bullet_mgr:
                if pygame.Rect.colliderect(pygame.Rect(enemy.rect.left - 5, enemy.rect.top - 5, enemy.rect.width + 10, enemy.rect.height + 10), bullet.rect):
                    bullet.done = True
                    enemy.done = True
        for enemy in self.enemies:
            if enemy.done:
                self.enemies.remove(enemy)
                self.win()
        if self.arrows == 0 and len(self.bullet_mgr.sprites()) == 0:
            self.die_arrows()


class Player(pygame.sprite.Sprite):
    
    def __init__(self, pos, bullet_mgr):
        pygame.sprite.Sprite.__init__(self)
        self.pos = list(pos)

        ss_image = pygame.image.load(data.filepath('character.png')).convert()
        ss_image.set_colorkey((255, 255, 255))
        spritesheet = SpriteSheetSlicer().slice(ss_image, const.TILE_SIZE, 2 * const.TILE_SIZE)
        self.walk_frame_interval = 0.2
        self.stand_right = spritesheet[0][0]
        self.walk_right = Animation([spritesheet[0][0], spritesheet[0][1]], self.walk_frame_interval)
        self.stand_left = spritesheet[1][0]
        self.walk_left = Animation([spritesheet[1][0], spritesheet[1][1]], self.walk_frame_interval)
        self.stand_down = spritesheet[0][2]
        self.walk_down = Animation([spritesheet[0][2], spritesheet[0][3]], self.walk_frame_interval)
        self.stand_up = spritesheet[1][2]
        self.walk_up = Animation([spritesheet[1][2], spritesheet[1][3]], self.walk_frame_interval)
        self.aiming_down = spritesheet[0][5]
        self.aiming_up = spritesheet[1][5]
        self.aiming_right = spritesheet[0][4]
        self.aiming_left = spritesheet[1][4]
        self.placing_mirror = spritesheet[0][6]
        self.image = self.stand_down
        self.direction = const.DOWN

        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.speed = 100
        self.vel = [0, 0]
        self.moving_left, self.moving_right, self.moving_up, self.moving_down = False, False, False, False
        self.action_direction = const.NODIR
        self.placing = False
        self.aiming = False
        self.bullet_mgr = bullet_mgr
        self.bullet_speed = 300
    
    def move_left(self): self.moving_left = True
    def move_right(self): self.moving_right = True
    def move_up(self): self.moving_up = True
    def move_down(self): self.moving_down = True
    def stop_move_left(self): self.moving_left = False
    def stop_move_right(self): self.moving_right = False
    def stop_move_up(self): self.moving_up = False
    def stop_move_down(self): self.moving_down = False
    
    def get_center(self): return (self.pos[0] + self.rect.width / 2,
                                  self.pos[1] + self.rect.height - 4)
    
    def update_move(self):
        if self.moving_left: self.vel[0] = -self.speed
        elif self.moving_right: self.vel[0] = self.speed
        else: self.vel[0] = 0
        if self.moving_up: self.vel[1] = -self.speed
        elif self.moving_down: self.vel[1] = self.speed
        else: self.vel[1] = 0

    def update_direction(self):
        if self.vel[1] > 0:
            self.direction = const.DOWN
            self.image = self.walk_down
        elif self.vel[1] < 0:
            self.direction = const.UP
            self.image = self.walk_up
        else:
            if self.vel[0] > 0:
                self.direction = const.RIGHT
                self.image = self.walk_right
            elif self.vel[0] < 0:
                self.direction = const.LEFT
                self.image = self.walk_left
            else: self.stop_image()

    def stop_image(self):
        if self.direction == const.DOWN: self.image = self.stand_down
        elif self.direction == const.UP: self.image = self.stand_up
        elif self.direction == const.RIGHT: self.image = self.stand_right
        elif self.direction == const.LEFT: self.image = self.stand_left
        
    def update(self, dt):
        if self.placing or self.aiming:
            self.update_action_direction()
            self.vel[0] = 0
            self.vel[1] = 0
            if self.placing:
                self.image = self.placing_mirror
        else:
            self.update_move()
            self.update_direction()
        self.pos[0] += self.vel[0] * dt
        self.pos[1] += self.vel[1] * dt
        if self.pos[0] <= 0:
            self.pos[0] = 0
        if self.pos[0] + self.rect.width >= const.WIDTH:
            self.pos[0] = const.WIDTH - self.rect.width
        if self.pos[1] <= 0:
            self.pos[1] = 0
        if self.pos[1] + self.rect.height >= const.HEIGHT:
            self.pos[1] = const.HEIGHT - self.rect.height
        self.rect.topleft = self.pos
        if isinstance(self.image, Animation):
            self.image.update(dt)
    
    def start_directed_action(self, aiming):
        if aiming: self.aiming = True
        else: self.placing = True
        self.update_action_direction()
    
    def update_action_direction(self):
        if self.moving_up: self.action_direction = const.UP
        elif self.moving_down: self.action_direction = const.DOWN
        elif self.moving_left: self.action_direction = const.LEFT
        elif self.moving_right: self.action_direction = const.RIGHT
        else: self.action_direction = const.NODIR
        if self.aiming:
            if self.moving_up: self.image = self.aiming_up
            elif self.moving_down: self.image = self.aiming_down
            elif self.moving_left: self.image = self.aiming_left
            elif self.moving_right: self.image = self.aiming_right
            else: self.image = self.stand_right
    
    def place_mirror_done(self, tilemap):
        self.placing = False

        tilepos = util.tilepos(*self.get_center())
        tile_obj = tilemap.objects[tilepos[0]][tilepos[1]]
        if tile_obj is not None and tile_obj.obj_type == 'mirror':
            tilemap.remove_object(tilepos)
        elif self.action_direction == const.NODIR:
            return
        
        on_moonlight = tilemap.lights[tilepos[0]][tilepos[1]] is not None
        on_ray = len(tilemap.rays[tilepos[0]][tilepos[1]]) > 0

        if tile_obj is None and (on_moonlight or on_ray):
            if on_ray:
                direction = tilemap.rays[tilepos[0]][tilepos[1]][0].direction
                if direction == const.UP and self.action_direction in [const.UP, const.DOWN]: return
                if direction == const.DOWN and self.action_direction in [const.UP, const.DOWN]: return
                if direction == const.LEFT and self.action_direction in [const.LEFT, const.RIGHT]: return
                if direction == const.RIGHT and self.action_direction in [const.LEFT, const.RIGHT]: return
            tilemap.add_object(Mirror(tilepos, self.action_direction, tilemap, on_moonlight, tilemap.rays[tilepos[0]][tilepos[1]][0].direction if on_ray else None))
        
        for mirror in tilemap.mirrors:
            mirror.update_path()
    
    def shoot_done(self, hunted):
        self.aiming = False
        if self.action_direction == const.NODIR: return False
        if hunted: return False

        if self.action_direction == const.UP: vel = [0, -self.bullet_speed]
        if self.action_direction == const.DOWN: vel = [0, self.bullet_speed]
        if self.action_direction == const.LEFT: vel = [-self.bullet_speed, 0]
        if self.action_direction == const.RIGHT: vel = [self.bullet_speed, 0]
        self.bullet_mgr.create_bullet(self.rect.center, vel)
        return True
    
    def get_bb(self):
        return pygame.rect.Rect(self.rect.left + 6, self.rect.top + const.TILE_SIZE, 7, 14)
        

class BulletManager(pygame.sprite.RenderUpdates):

    def __init__(self):
        pygame.sprite.RenderUpdates.__init__(self)

    def create_bullet(self, pos, vel):
        bullet = Bullet(pos, vel)
        self.add(bullet)
    
    def update(self, dt):
        for bullet in self:
            bullet.update(dt)
            if bullet.done:
                self.remove(bullet)


class Bullet(pygame.sprite.Sprite):

    def __init__(self, pos, vel):
        pygame.sprite.Sprite.__init__(self)
        self.pos = list(pos)
        self.vel = list(vel)
        if vel[1] == 0:
            self.image = pygame.surface.Surface((3, 1))
            self.image.fill((153,123,26))
            self.image.set_at((2,0), (200,200,200))
        else:
            self.image = pygame.surface.Surface((1, 3))
            self.image.fill((153,123,26))
            self.image.set_at((0,2), (200,200,200))
        self.rect = self.image.get_rect()
        self.rect.center = self.pos
        self.done = False
        self.timer = 0.75

    def update(self, dt):
        self.timer -= dt
        self.pos[0] += self.vel[0] * dt
        self.pos[1] += self.vel[1] * dt
        self.rect.center = self.pos
        if self.timer <= 0 or (self.pos[0] + self.rect.width < 0 or self.pos[0] > const.WIDTH or
                               self.pos[1] + self.rect.height < 0 or self.pos[1] > const.HEIGHT):
            self.done = True


class Tilemap(object):
    
    def __init__(self):
        self.map = [[None for i in range(const.TILES_X)] for j in range(const.TILES_Y)]
        self.lights = [[None for i in range(const.TILES_X)] for j in range(const.TILES_Y)]
        self.objects = [[None for i in range(const.TILES_X)] for j in range(const.TILES_Y)]
        self.rays = [[[] for i in range(const.TILES_X)] for j in range(const.TILES_Y)]
        self.map_sprites = pygame.sprite.RenderUpdates()
        self.light_sprites = pygame.sprite.RenderUpdates()
        self.object_sprites = pygame.sprite.RenderUpdates()
        self.mirrors = []
    
    def add_tile(self, tile):
        if self.map[tile.tilepos[0]][tile.tilepos[1]] is not None:
            return False
        self.map_sprites.add(tile)
        self.map[tile.tilepos[0]][tile.tilepos[1]] = tile
        return True
    
    def add_light(self, tile):
        if self.lights[tile.tilepos[0]][tile.tilepos[1]] is not None:
            return False
        self.light_sprites.add(tile)
        self.lights[tile.tilepos[0]][tile.tilepos[1]] = tile
        return True
    
    def add_object(self, object):
        if self.objects[object.tilepos[0]][object.tilepos[1]] is not None:
            return False
        self.object_sprites.add(object)
        self.objects[object.tilepos[0]][object.tilepos[1]] = object
        if object.obj_type == 'mirror':
            self.mirrors.append(object)
        return True
    
    def add_ray_tile(self, ray, tilepos):
        self.rays[tilepos[0]][tilepos[1]].append(ray)
    
    def clear_ray(self, ray):
        for i in range(const.TILES_Y):
            for j in range(const.TILES_X):
                for ray_tile in [ray_tile for ray_tile in self.rays[i][j] if ray_tile.ray == ray]:
                    self.rays[i][j].remove(ray_tile)
    
    def remove_object(self, tilepos):
        obj = self.objects[tilepos[0]][tilepos[1]]
        if obj is None: return
        self.objects[tilepos[0]][tilepos[1]] = None
        self.object_sprites.remove(obj)
        if obj.obj_type == 'mirror':
            self.mirrors.remove(obj)
    
    def update(self, dt):
        pass


class Tile(pygame.sprite.Sprite):

    def __init__(self, tilepos):
        pygame.sprite.Sprite.__init__(self)
        self.pos = list(util.pos(*tilepos))
        self.tilepos = tilepos
        self.image = pygame.Surface((const.TILE_SIZE,const.TILE_SIZE))
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.tile_type = 'none'


class TileBush(Tile):
    
    def __init__(self, tilepos):
        Tile.__init__(self, tilepos)
        self.image = pygame.image.load(data.filepath('bush.png')).convert()
        self.tile_type = 'bush'


class TileGrass(Tile):

    def __init__(self, tilepos):
        Tile.__init__(self, tilepos)
        self.image = pygame.image.load(data.filepath('grass.png')).convert()
        self.tile_type = 'grass'


class TileMoonlight(Tile):
    
    def __init__(self, tilepos):
        Tile.__init__(self, tilepos)
        self.image = pygame.image.load(data.filepath('moonlight.png')).convert()
        self.image.set_colorkey((255,255,255))
        self.tile_type = 'moonlight'


class TileObject(pygame.sprite.Sprite):

    def __init__(self, tilepos):
        pygame.sprite.Sprite.__init__(self)
        self.pos = util.pos(*tilepos)
        self.tilepos = tilepos
        self.image = pygame.Surface((const.TILE_SIZE,const.TILE_SIZE))
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.obj_type = 'none'


class Mirror(TileObject):

    def __init__(self, tilepos, direction, tilemap, onlight, prev_dir):
        TileObject.__init__(self, tilepos)
        ss_image = pygame.image.load(data.filepath('mirrors.png')).convert()
        ss_image.set_colorkey((255, 255, 255))
        spritesheet = SpriteSheetSlicer().slice(ss_image, const.TILE_SIZE, const.TILE_SIZE)
        self.direction = direction
        if onlight:
            if direction == const.UP:
                self.image = spritesheet[0][1]
            elif direction == const.DOWN:
                self.image = spritesheet[0][0]
            elif direction == const.LEFT:
                self.image = spritesheet[1][0]
            else:
                self.image = spritesheet[1][1]
        else:
            if (direction == const.UP and prev_dir == const.RIGHT) or (direction == const.LEFT and prev_dir == const.DOWN):
                self.image = spritesheet[1][3]
            if (direction == const.UP and prev_dir == const.LEFT) or (direction == const.RIGHT and prev_dir == const.DOWN):
                self.image = spritesheet[1][2]
            if (direction == const.DOWN and prev_dir == const.RIGHT) or (direction == const.LEFT and prev_dir == const.UP):
                self.image = spritesheet[0][2]
            if (direction == const.DOWN and prev_dir == const.LEFT) or (direction == const.RIGHT and prev_dir == const.UP):
                self.image = spritesheet[0][3]
        self.ray = None
        if onlight:
            self.ray = Ray(self.direction, self.tilepos, tilemap)
        self.obj_type = 'mirror'
    
    def update_path(self):
        if self.ray is not None: self.ray.update_path()


class Ray(pygame.sprite.RenderUpdates):
    
    def __init__(self, direction, origin, tilemap):
        pygame.sprite.RenderUpdates.__init__(self)
        self.direction = direction
        self.origin = origin
        self.length = 25
        self.tilemap = tilemap
        self.compute_path()
    
    def compute_path(self):
        self.empty()
        first_ray = True
        self.path = []
        current = self.origin
        cur_dir = self.direction
        for i in range(self.length):
            obj = self.tilemap.objects[current[0]][current[1]]
            end_ray = False
            if obj is not None and obj.obj_type == 'mirror':
                end_ray = True
            ray_tile = RayTile(current, self, cur_dir, first_ray, end_ray)
            self.add(ray_tile)
            self.path.append(ray_tile)
            self.tilemap.add_ray_tile(ray_tile, current)
            if obj is not None and obj.obj_type == 'mirror':
                cur_dir = obj.direction
                ray_tile = RayTile(current, self, cur_dir, True, False)
                self.add(ray_tile)
                self.path.append(ray_tile)
                self.tilemap.add_ray_tile(ray_tile, current)
            current = util.next_tile(current, cur_dir)
            first_ray = False
            if current is None:
                break
    
    def update_path(self):
        self.tilemap.clear_ray(self)
        self.compute_path()
    
    def update(self, dt):
        pass


class RayTile(Tile):
    
    def __init__(self, tilepos, ray, direction, half, half_end):
        Tile.__init__(self, tilepos)
        if half:
            self.ray_length = const.TILE_SIZE / 2
            if direction == const.RIGHT:
                self.pos[0] += const.TILE_SIZE / 2
            if direction == const.DOWN:
                self.pos[1] += const.TILE_SIZE / 2
        elif half_end:
            self.ray_length = const.TILE_SIZE / 2
            if direction == const.LEFT:
                self.pos[0] += const.TILE_SIZE / 2
            if direction == const.UP:
                self.pos[1] += const.TILE_SIZE / 2
        else:
            self.ray_length = const.TILE_SIZE
        self.half = half
        self.ray_width = 4
        self.cut_off = True
        self.direction = direction
        self.restore()
        self.ray = ray
        self.appear()
    
    def shorten(self, new_size):
        self.cut_off = True
        if self.direction == const.RIGHT:
            self.image = pygame.surface.Surface((new_size, self.ray_width))
            self.image.fill((255,255,255))
        if self.direction == const.LEFT:
            self.image = pygame.surface.Surface((self.ray_length, self.ray_width))
            self.image.fill((255,255,255))
            self.image.set_colorkey((255,0,255))
            if self.ray_length - new_size > 0:
                empty_surface = pygame.surface.Surface((self.ray_length - new_size, self.ray_width))
                empty_surface.fill((255,0,255))
                self.image.blit(empty_surface, (0,0))
        if self.direction == const.DOWN:
            self.image = pygame.surface.Surface((self.ray_width, new_size))
            self.image.fill((255,255,255))
        if self.direction == const.UP:
            self.image = pygame.surface.Surface((self.ray_width, self.ray_length))
            self.image.fill((255,255,255))
            self.image.set_colorkey((255,0,255))
            if self.ray_length - new_size > 0:
                empty_surface = pygame.surface.Surface((self.ray_width, self.ray_length - new_size))
                empty_surface.fill((255,0,255))
                self.image.blit(empty_surface, (0,0))
    
    def restore(self):
        if self.cut_off:
            if self.direction in [const.LEFT, const.RIGHT]:
                self.image = pygame.surface.Surface((self.ray_length, self.ray_width))
                self.rect = self.image.get_rect()
                self.rect.topleft = (self.pos[0], self.pos[1] + const.TILE_SIZE / 2 - self.ray_width / 2)
            else:
                self.image = pygame.surface.Surface((self.ray_width, self.ray_length))
                self.rect = self.image.get_rect()
                self.rect.topleft = (self.pos[0] + const.TILE_SIZE / 2 - self.ray_width / 2, self.pos[1])
            self.image.fill((255,255,255))
            self.cut_off = False
    
    def disappear(self):
        self.image.set_alpha(0)

    def appear(self):
        self.image.set_alpha(64)


class RayBlockManager(object):
    
    def __init__(self, player, enemies, tilemap):
        self.player = player
        self.enemies = enemies
        self.tilemap = tilemap
    
    def update(self, dt):
        for mirror in self.tilemap.mirrors:
            if mirror.ray is None: continue
            remove = False
            for ray_tile in mirror.ray.path:
                if remove:
                    ray_tile.disappear()
                else:
                    for rect in [self.player.get_bb()] + [enemy.rect for enemy in self.enemies]:
                        if ray_tile.direction == const.RIGHT:
                            if pygame.Rect.colliderect(pygame.Rect(rect.left, rect.top, 1, rect.height), ray_tile.rect):
                                ray_tile.shorten(rect.left % const.TILE_SIZE)
                                remove = True
                        if ray_tile.direction == const.LEFT:
                            if pygame.Rect.colliderect(pygame.Rect(rect.right - 1, rect.top, 2, rect.height), ray_tile.rect):
                                ray_tile.shorten(const.TILE_SIZE - rect.right % const.TILE_SIZE)
                                remove = True
                        if ray_tile.direction == const.DOWN:
                            if pygame.Rect.colliderect(pygame.Rect(rect.left, rect.top, rect.width, 1), ray_tile.rect):
                                ray_tile.shorten(rect.top % const.TILE_SIZE)
                                remove = True
                        if ray_tile.direction == const.UP:
                            if pygame.Rect.colliderect(pygame.Rect(rect.left, rect.bottom - 1, rect.width, 2), ray_tile.rect):
                                ray_tile.shorten(const.TILE_SIZE - rect.bottom % const.TILE_SIZE)
                                remove = True
                    if not remove: ray_tile.restore()
                    ray_tile.appear()


class Enemy(pygame.sprite.Sprite):
    
    def __init__(self, pos, player):
        pygame.sprite.Sprite.__init__(self)
        self.pos = list(pos)
        self.image = pygame.image.load(data.filepath('enemy.png')).convert()
        self.image.set_colorkey((255,255,255))
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.vel = [0, 0]
        self.speed = 40
        self.state = const.STATE_WANDER
        self.player = player
        self.wander_timer = 0
        self.wander_direction = const.NODIR
        self.sight_range = 60
        self.hunt_range = 100
        self.done = False
    
    def update(self, dt):
        self.check_player()
        if self.state == const.STATE_WANDER:
            self.wander_timer -= dt
            if self.wander_timer <= 0:
                self.wander_timer = random.uniform(1, 6)
                self.wander_direction = random.choice(const.DIRSALL)
                if self.wander_direction == const.LEFT:
                    self.vel = [-self.speed, 0]
                elif self.wander_direction == const.RIGHT:
                    self.vel = [self.speed, 0]
                elif self.wander_direction == const.UP:
                    self.vel = [0, -self.speed]
                elif self.wander_direction == const.DOWN:
                    self.vel = [0, self.speed]
                else:
                    self.vel = [0, 0]
            self.pos[0] += self.vel[0] * dt
            self.pos[1] += self.vel[1] * dt
        elif self.state == const.STATE_HUNT:
            dx = self.player.pos[0] - self.pos[0]
            dy = self.player.pos[1] - self.pos[1]
            dist = math.hypot(dx, dy)
            if dist <= self.speed * dt:
                self.pos[0] = self.player.pos[0]
                self.pos[1] = self.player.pos[1]
            else:
                self.pos[0] += self.speed * dt * dx / dist
                self.pos[1] += self.speed * dt * dy / dist
        if self.pos[0] <= 0:
            self.pos[0] = 0
            if self.state == const.STATE_WANDER: self.wander_timer = 0
        if self.pos[0] + self.rect.width >= const.WIDTH:
            self.pos[0] = const.WIDTH - self.rect.width
            if self.state == const.STATE_WANDER: self.wander_timer = 0
        if self.pos[1] <= 0:
            self.pos[1] = 0
            if self.state == const.STATE_WANDER: self.wander_timer = 0
        if self.pos[1] + self.rect.height >= const.HEIGHT:
            self.pos[1] = const.HEIGHT - self.rect.height
            if self.state == const.STATE_WANDER: self.wander_timer = 0
        self.rect.topleft = self.pos
    
    def check_player(self):
        dx = self.player.pos[0] - self.pos[0]
        dy = self.player.pos[1] - self.pos[1]
        dist = math.hypot(dx, dy)
        dist_range = self.sight_range if self.state == const.STATE_WANDER else self.hunt_range
        if dist <= dist_range:
            self.state = const.STATE_HUNT
        else:
            self.state = const.STATE_WANDER


class DamageScreen(pygame.sprite.Sprite):
    
    def __init__(self):
        pygame.sprite.Sprite.__init__(self)
        self.pos = [0,0]
        self.image = pygame.image.load(data.filepath('damage.png')).convert()
        self.image.set_colorkey((255,255,255))
        self.image.set_alpha(128)
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.active = False
        self.timer = 0.2
        self.callback = None
    
    def start(self, callback):
        self.active = True
        self.timer = 0.2
        self.callback = callback
    
    def update(self, dt):
        self.timer -= dt
        if self.timer <= 0:
            self.active = False
            if self.callback is not None: self.callback()


class HuntedScreen(pygame.sprite.Sprite):

    def __init__(self):
        pygame.sprite.Sprite.__init__(self)
        self.pos = [0,0]
        self.image = pygame.image.load(data.filepath('hunted.png')).convert()
        self.image.set_colorkey((255,255,255))
        self.image.set_alpha(0)
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.hunted = False
    
    def set_active(self, enable):
        self.image.set_alpha(64 if enable else 0)
        self.hunted = enable


class Sight(pygame.sprite.Sprite):

    def __init__(self, player):
        pygame.sprite.Sprite.__init__(self)
        self.player = player
        self.image = pygame.image.load(data.filepath('sight.png')).convert_alpha()
        self.rect = self.image.get_rect()
        self.rect.center = self.player.get_bb().center

    def update(self, dt):
        self.rect.center = self.player.get_bb().center


class SpriteSheetSlicer(object):

    def slice(self, image, tile_size_x, tile_size_y):
        image_width, image_height = image.get_size()
        width = int(image_width / tile_size_x)
        height = int(image_height / tile_size_y)
        images = [[None for j in xrange(width)] for i in xrange(height)]
        for i in xrange(width):
            for j in xrange(height):
                images[j][i] = image.subsurface(i * tile_size_x, j * tile_size_y,
                    tile_size_x, tile_size_y)
        return images


class Animation(pygame.surface.Surface):

    def __init__(self, frames, interval):
        pygame.surface.Surface.__init__(self, frames[0].get_size())
        self.frames = frames
        self.current_frame = 0
        self.interval = interval
        self.timer = interval
        self.clear()
        self.blit(self.frames[self.current_frame], (0, 0))

    def clear(self):
        self.fill((255, 0, 255))
        self.set_colorkey((255, 0, 255))

    def update(self, dt):
        self.timer -= dt
        if self.timer <= 0:
            self.timer = self.interval
            self.current_frame += 1
            if self.current_frame >= len(self.frames):
                self.current_frame = 0
            self.clear()
            self.blit(self.frames[self.current_frame], (0, 0))
